# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import asyncio
import os

try:
    import pty
    has_pty = True
except ImportError:
    has_pty = False

from ..get_loop_impl import get_loop_impl


def get_loop():
    return get_loop_impl(asyncio)


async def _async_execute_process_nopty(
    protocol_class, cmd, cwd, env, shell,
    stderr_to_stdout=True
):
    loop = get_loop()
    stderr = asyncio.subprocess.PIPE
    if stderr_to_stdout is True:
        stderr = asyncio.subprocess.STDOUT
    # Start the subprocess
    if shell is True:
        transport, protocol = await loop.subprocess_shell(
            protocol_class, " ".join(cmd), cwd=cwd, env=env,
            stderr=stderr, close_fds=False)
    else:
        transport, protocol = await loop.subprocess_exec(
            protocol_class, *cmd, cwd=cwd, env=env,
            stderr=stderr, close_fds=False)
    return transport, protocol


if has_pty:
    # If pty is availabe, use them to emulate the tty
    async def _async_execute_process_pty(
        protocol_class, cmd, cwd, env, shell,
        stderr_to_stdout=True
    ):
        loop = get_loop()
        # Create the PTY's
        stdout_master, stdout_slave = pty.openpty()
        if stderr_to_stdout:
            stderr_master, stderr_slave = stdout_master, stdout_slave
        else:
            stderr_master, stderr_slave = pty.openpty()

        def protocol_factory():
            return protocol_class(
                stdin=None,
                stdout=stdout_master,
                stderr=stderr_master
            )

        # Start the subprocess
        if shell is True:
            transport, protocol = await loop.subprocess_shell(
                protocol_factory, " ".join(cmd), cwd=cwd, env=env,
                stdout=stdout_slave, stderr=stderr_slave, close_fds=False)
        else:
            transport, protocol = await loop.subprocess_exec(
                protocol_factory, *cmd, cwd=cwd, env=env,
                stdout=stdout_slave, stderr=stderr_slave, close_fds=False)

        # Close our copies of the slaves,
        # the child's copy of the slave remain open until it terminates
        os.close(stdout_slave)
        if not stderr_to_stdout:
            os.close(stderr_slave)

        # Create Protocol classes
        class PtyStdoutProtocol(asyncio.Protocol):
            def connection_made(self, transport):
                if hasattr(protocol, 'on_stdout_open'):
                    protocol.on_stdout_open()

            def data_received(self, data):
                if hasattr(protocol, 'on_stdout_received'):
                    protocol.on_stdout_received(data)

            def connection_lost(self, exc):
                if hasattr(protocol, 'on_stdout_close'):
                    protocol.on_stdout_close(exc)

        class PtyStderrProtocol(asyncio.Protocol):
            def connection_made(self, transport):
                if hasattr(protocol, 'on_stderr_open'):
                    protocol.on_stderr_open()

            def data_received(self, data):
                if hasattr(protocol, 'on_stderr_received'):
                    protocol.on_stderr_received(data)

            def connection_lost(self, exc):
                if hasattr(protocol, 'on_stderr_close'):
                    protocol.on_stderr_close(exc)

        # Add the pty's to the read loop
        # Also store the transport, protocol tuple for each call to
        # connect_read_pipe, to prevent the destruction of the protocol
        # class instance, otherwise no data is received.
        protocol.stdout_tuple = await loop.connect_read_pipe(
            PtyStdoutProtocol, os.fdopen(stdout_master, 'rb', 0))
        if not stderr_to_stdout:
            protocol.stderr_tuple = await loop.connect_read_pipe(
                PtyStderrProtocol, os.fdopen(stderr_master, 'rb', 0))
        # Return the protocol and transport
        return transport, protocol
else:
    _async_execute_process_pty = _async_execute_process_nopty


async def async_execute_process(
    protocol_class, cmd=None, cwd=None, env=None, shell=False,
    emulate_tty=False, stderr_to_stdout=True
):
    if emulate_tty:
        transport, protocol = await _async_execute_process_pty(
            protocol_class, cmd, cwd, env, shell,
            stderr_to_stdout)
    else:
        transport, protocol = await _async_execute_process_nopty(
            protocol_class, cmd, cwd, env, shell,
            stderr_to_stdout)
    return transport, protocol



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module provides functions for doing process management.

The documentation for this module has a custom layout in process_utils.rst.
"""

from .async_execute_process import async_execute_process
from .async_execute_process import asyncio
from .async_execute_process import AsyncSubprocessProtocol
from .async_execute_process import get_loop

from .impl import execute_process
from .impl import execute_process_split
from .impl import which

__all__ = [
    'async_execute_process',
    'asyncio',
    'AsyncSubprocessProtocol',
    'get_loop',
    'execute_process',
    'execute_process_split',
    'which',
]



# Copyright 2014 Open Source Robotics Foundation, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
This module has a miscellaneous set of functions for working with terminals.

You can use the :py:func:`get_terminal_dimensions` to get the width and
height of the terminal as a tuple.

You can also use the :py:func:`is_tty` function to determine if a given object
is a tty.
"""

import os
import struct
import subprocess

__all__ = ['GetTerminalDimensionsError', 'get_terminal_dimensions', 'is_tty']


class GetTerminalDimensionsError(Exception):
    """Raised when the terminal dimensions cannot be determined."""
    pass


def _get_terminal_dimensions_windows():
    try:
        from ctypes import create_string_buffer
        from ctypes import windll
    except ImportError as exc:
        raise GetTerminalDimensionsError("Failed to get dimensions: {0}"
                                         .format(exc))
    STDOUT = -11
    h = windll.kernel32.GetStdHandle(STDOUT)
    buffer_info = create_string_buffer(22)
    if not windll.kernel32.GetConsoleScreenBufferInfo(h, buffer_info):
        raise GetTerminalDimensionsError(
            "Call to windll.kernel32.GetConsoleScreenBufferInfo failed")

    try:
        (_, _, _, _, _, left, top, right, bottom, _, _) \
            = struct.unpack("hhhhHhhhhhh", buffer_info.raw)
    except struct.error as exc:
        raise GetTerminalDimensionsError("Failed to unpack data: {0}"
                                         .format(exc))
    width = right - left + 1
    height = bottom - top + 1

    return width, height


def _get_terminal_dimensions_unix():
    # This function uses `tput` and should work on any Unix system
    # See: http://en.wikipedia.org/wiki/Tput
    try:
        width = subprocess.check_output(['tput', 'cols'])

        width = int(width.strip())
    except (subprocess.CalledProcessError, ValueError) as exc:
        raise GetTerminalDimensionsError("Failed to get width: {0}"
                                         .format(exc))
    try:
        height = subprocess.check_output(['tput', 'lines'])

        height = int(height.strip())
    except (subprocess.CalledProcessError, ValueError) as exc:
        raise GetTerminalDimensionsError("Failed to get height: {0}"
                                         .format(exc))
    return width, height


def get_terminal_dimensions():
    """Returns the width and height of the terminal.

    :returns: width and height in that order as a tuple
    :rtype: tuple
    :raises: GetTerminalDimensionsError when the terminal dimensions
        cannot be determined
    """
    if os.name in ['nt']:
        return _get_terminal_dimensions_windows()
    return _get_terminal_dimensions_unix()


def is_tty(stream):
    """Returns True if the given stream is a tty, else False

    :param stream: object to be checked for being a tty
    :returns: True if the given object is a tty, otherwise False
    :rtype: bool
    """
    return hasattr(stream, 'isatty') and stream.isatty()



from setuptools import find_packages
from setuptools import setup


install_requires = [
    'setuptools',
]
package_excludes = ['tests*', 'docs*']
packages = find_packages(exclude=package_excludes)

package_name = 'osrf_pycommon'

setup(
    name=package_name,
    version='1.0.3',
    packages=packages,
    data_files=[
        ('share/' + package_name, ['package.xml']),
        ('share/ament_index/resource_index/packages',
            ['resource/' + package_name]),
    ],
    install_requires=install_requires,
    python_requires='>=3.5',
    zip_safe=True,
    author='William Woodall',
    author_email='william@osrfoundation.org',
    maintainer='William Woodall',
    maintainer_email='william@osrfoundation.org',
    url='http://osrf-pycommon.readthedocs.org/',
    keywords=['osrf', 'utilities'],
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python',
    ],
    description="Commonly needed Python modules, "
                "used by Python software developed at OSRF",
    license='Apache 2.0',
    test_suite='tests',
)
